
// Series RLC circuit AC analysis
// ==============================
// Frequency response of series RLC circuit
// Method: Runge-Kutta 4
// Needs linking with math library (GCC option -lm)
// Warning: not validated
// Copyright: (c) Petr Peringer, FIT TU Brno, 2017,2018

// Circuit schematics:
//           _____
//     o----|_____|---*------o
//            Rg      |
//                   | |
//                   | | R
//                   | |
//                    |
//                     )
//   Uin               ) L       Uout
//                     )
//                    |
//                    |
//                   === C
//                    |
//     o--------------*------o
//

#include <stdio.h>
#include <math.h>               // sin(), M_PI

// System: series RLC
#define SIZE 2                  // State vector size of the system
// model parameters:
      double R = 10;            // 10 ohm, will be changed
const double L = 100e-3;        // 100 uF
const double C = 100e-6;        // 100 mH
// input:
const double Rg = 1000;         // Source impedance
const double A  = 1.0;          // Amplitude (NOT effective value)
      double freq = 50;         // center frequency

// compute input voltage
inline double uin(double t) {
    return A*sin(2*M_PI*freq*t); // sin(\omega t)
}

// compute output voltage
inline double uout(double t, double y[]) {
    // state variables:
    double u_C = y[0];
    double i   = y[1];
    // algebraic equations (converted):
    double u_Rg= Rg * i;        // signal generator
    double u_R = R * i;         // Ohm's law
    double u_L = uin(t) - u_Rg - u_R - u_C;
    // output voltage:
    return u_R + u_L + u_C;
}

// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    // state variables:
    double u_C = y[0];
    double i   = y[1];
    // algebraic equations (converted):
    double u_Rg= Rg * i;        // signal generator voltage
    double u_R = R * i;         // Ohm's law
    double u_L = uin(t) - u_Rg - u_R - u_C; // Kirchhoff II
    // differential equations:
    yin[0] = i/C;               // der(u_C)
    yin[1] = u_L/L;             // der(i)
}

// Runge-Kutta-4 step
void RK4step(double t, double h, double *y);

// Experiment control
void Run()
{
    // Simulation parameters:
    const double FRes = 1/(2*M_PI*sqrt(L*C)); // resonant frequency
    const double TRes = 1/FRes;               // period
    const double stepsize = TRes/300;         // RK4 step size
    printf("# RLC analysis\n");
    printf("# R=%g\n",R);
    printf("# L=%g\n",L);
    printf("# C=%g\n",C);
    printf("# Rg=%g\n",Rg);
    printf("# FRes=%g\n",FRes);
    double q=1.1;
    for(freq=FRes/15;freq<FRes*15;freq*=q) {    // change input frequency
        q=(fabs(freq-FRes)/FRes < 0.2)?1.005:1.1;
        const double T    = 1/freq;     // period of input signal
        const double TEND = 50*T;       // how many periods
        double y[SIZE] = { 0, 0 };      // initial values
        double umax=0;                  // amplitude
        double t = 0;                   // time
        while (t < TEND) {
            if(t>0.9*TEND && uout(t,y)>umax)
                 umax=uout(t,y);
            RK4step(t, stepsize, y);    // make step
            t += stepsize;              // advance the simulation time
        }
        printf("%8.3f  %8f\n", freq, umax);
    }
    printf("\n");
}

int main()
{
    R = 10;
    Run();      // experiment1
    R = 1;
    Run();      // experiment2
    return 0;
}

//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ystart[SIZE];        // initial state
    // 4 stages results:
    double k1[SIZE];
    double k2[SIZE];
    double k3[SIZE];
    double k4[SIZE];

    int i;
    for (i = 0; i < SIZE; i++)  // save initial value
        ystart[i] = y[i];

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        k1[i] = h * yin[i];     // k1 = h * f(t, y(t))
        y[i] = ystart[i] + k1[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        k2[i] = h * yin[i];     // k2 = h * f(t+h/2, y(t)+k1/2)
        y[i] = ystart[i] + k2[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        k3[i] = h * yin[i];     // k3 = h * f(t+h/2, y(t)+k2/2)
        y[i] = ystart[i] + k3[i];
    }

    f(t + h, y, yin);           // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        k4[i] = h * yin[i];     // k4 = h * f(t+h, y(t)+k3)
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] = ystart[i] + k1[i] / 6 + k2[i] / 3 + k3[i] / 3 + k4[i] / 6;
    }
    // Return: y = new state at time t+h
}
